package de.uni.leipzig.informatik.pacosy.cophylogeny.main;

import ilog.concert.IloException;

import java.io.File;
import java.io.IOException;
import java.util.Comparator;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.GnuParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.Options;

import de.uni.leipzig.informatik.pacosy.cophylogeny.Parameters;
import de.uni.leipzig.informatik.pacosy.cophylogeny.solver.cplex.CplexCophylogenySolver;
import de.uni.leipzig.informatik.pacosy.datastructure.dao.Tanglegramm;
import de.uni.leipzig.informatik.pacosy.nexus.Nexus;
import de.uni.leipzig.informatik.pacosy.nexus.block.NexusCophylogenyBlock;
import de.uni.leipzig.informatik.pacosy.nexus.block.NexusTaxaBlock;
import de.uni.leipzig.informatik.pacosy.nexus.block.NexusTreesBlock;
import de.uni.leipzig.informatik.pacosy.nexus.exception.NexusFormatException;

public class Main {

	/**
	 * @param args
	 * @throws IOException 
	 * @throws IloException 
	 */
	public static void main(String[] args) throws IOException, IloException {
		Parameters parameters = Main.parseParameters(args);
		
		String filename = parameters.getInput();
		Nexus nexus = new Nexus(filename);
		NexusTaxaBlock taxaBlock = new NexusTaxaBlock();
		NexusTreesBlock treesBlock = new NexusTreesBlock();
		NexusCophylogenyBlock cophylogenyBlock = new NexusCophylogenyBlock();
		nexus.addNexusBlock("TAXA", taxaBlock);
		nexus.addNexusBlock("TREES", treesBlock);
		nexus.addNexusBlock("COPHYLOGENY", cophylogenyBlock);
		
		try {
			nexus.read();
		} catch (NexusFormatException e) {
			e.printStackTrace();
		}

		Tanglegramm tanglegramm = (Tanglegramm)cophylogenyBlock.getTanglegramm();
		CplexCophylogenySolver solver = new CplexCophylogenySolver(tanglegramm, parameters);
		solver.solve();
	}

	public static Options createOptions(){
		// create the Options
		Option h = new Option( "h", "help", false, "print this message");
		Option i = new Option( "i", "input", true, "input filename" );
		i.setRequired(true);
		Option o = new Option( "o", "output", true, "output filename" );
		o.setRequired(true);
		Option l = new Option( "l", "lower-bound", true, "lower bound for branch lenght factor [default=0.25]" );
		l.setType(Number.class);
		Option u = new Option( "u", "upper-bound", true, "upper bound for branch lenght factor [default=4.0]" );
		u.setType(Number.class);
		Option n = new Option( "n", "number-solutions", true, "maximal number of solutions [default=1]" );
		n.setType(Number.class);
		Option t = new Option( "t", "no-hostswitch", false, "do not minimize host switching events as second optimization creterion" );
		Option a = new Option( "a", "no-association", false, "do not minimize total associations as third optimization creterion" );
		Option s = new Option( "s", "single", false, "restrict solutions to only contail single associations" );
		Option v = new Option( "v", "verbose", false, "write ILP output to console" );
		
		Options options = new Options();
		options.addOption(h);
		options.addOption(i);
		options.addOption(o);
		options.addOption(l);
		options.addOption(u);
		options.addOption(n);
		options.addOption(t);
		options.addOption(a);
		options.addOption(s);
		options.addOption(v);
		
		return options;
	}
	
	public static Parameters parseParameters(String[] args){
		
		// create the command line parser
		CommandLineParser parser = new GnuParser();
		Options options = createOptions();
		Parameters parameters = new Parameters();
				
		try {
		    // parse the command line arguments
		    CommandLine line = parser.parse( options, args );
		    
		    if( line.hasOption( "help" ) ) {
		    	showHelp(options);
		    }

		    if( line.hasOption( "input" ) ) {
		    	String input = line.getOptionValue("input");
		    	parameters.setInput(input);
		    	if (input == "" || !new File(input).isFile()){
		    		System.out.println( "ERROR: input file '" + input + "' not found" );
		    		showHelp(options);
		    	}
		    }
		    
		    if( line.hasOption( "output" ) ) {
		    	String output = line.getOptionValue("output");
		    	parameters.setOutput(output);
		    	try {
					new File(output).getCanonicalPath();
				} catch (IOException e) {
					System.out.println( "ERROR: no valid output file specified" );
					showHelp(options);
				}
		    }
		    
		    if( line.hasOption( "lower-bound" ) ) {
		    	double lowerBound = ((Number)line.getParsedOptionValue("lower-bound")).doubleValue();
		    	parameters.setLowerBound(lowerBound);
		    }
		    
		    if( line.hasOption( "upper-bound" ) ) {
		    	double upperBound = ((Number)line.getParsedOptionValue("upper-bound")).doubleValue();
		    	parameters.setUpperBound(upperBound);
		    }
		    
		    if( line.hasOption( "number-solutions" ) ) {
		    	int noSolutions = ((Number)line.getParsedOptionValue("number-solutions")).intValue();
		    	parameters.setNoSolutions(noSolutions);
		    }
		    
		    if( line.hasOption( "no-hostswitch" ) ) {
		    	parameters.setHostswitch(false);
		    }
		    
		    if( line.hasOption( "no-association" ) ) {
		    	parameters.setAssociations(false);
		    }
		    
		    if( line.hasOption( "single" ) ) {
		    	parameters.setSingle(true);
		    }
		    
		    if( line.hasOption( "verbose" ) ) {
		    	boolean verbose = true;
		    	parameters.setVerbose(verbose);
		    }
		    
		}
		catch( Exception e ) {
		    System.out.println( e.getMessage() );
		    showHelp(options);
		}
		
		return parameters;
	}
	
	public static <T extends Option> void showHelp(Options options){
		HelpFormatter formatter = new HelpFormatter();
		formatter.setOptionComparator(new Comparator<T>() {
			private static final String OPTS_ORDER = "hioluntasv";
		    public int compare(T o1, T o2) {
		        return OPTS_ORDER.indexOf(o1.getOpt()) - OPTS_ORDER.indexOf(o2.getOpt());
		    }
		});
		formatter.printHelp( "java -jar CoRe-ILP.jar -i <input file> -o <output file> [OPTIONS]", options );
		System.exit(0);
	}
}
