package de.uni.leipzig.informatik.pacosy.nexus.command.parser;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import de.uni.leipzig.informatik.pacosy.datastructure.api.IEdge;
import de.uni.leipzig.informatik.pacosy.datastructure.api.INode;
import de.uni.leipzig.informatik.pacosy.datastructure.api.ITree;
import de.uni.leipzig.informatik.pacosy.datastructure.dao.factory.TreesFactory;
import de.uni.leipzig.informatik.pacosy.nexus.Token;
import de.uni.leipzig.informatik.pacosy.nexus.block.NexusBlock;
import de.uni.leipzig.informatik.pacosy.nexus.command.NexusCommand;
import de.uni.leipzig.informatik.pacosy.nexus.command.NexusMapOfStringsStyleCommand;
import de.uni.leipzig.informatik.pacosy.nexus.command.NexusTreeCommand;
import de.uni.leipzig.informatik.pacosy.nexus.exception.NexusFormatException;
import de.uni.leipzig.informatik.pacosy.util.Tupel;

public class NexusTreeCommandParser extends DefaultNexusCommandParser {

	protected static NexusTreeCommandParser singleton;
	
	public static NexusTreeCommandParser getSingleton(){
		if (singleton == null){
			singleton = new NexusTreeCommandParser();
		}
		return singleton;
	}
	
	private NexusTreeCommandParser() {}
	
	private int nextNodeId;
	private Map<String, INode> nodes;
	private Map<Tupel<String, String>, IEdge> edges;
	private Map<String, String> translate;

	@Override
	public TreesFactory getDaoFactory() {
		return (TreesFactory) super.getDaoFactory();
	}

	@Override
	public NexusTreeCommand deserialize(NexusCommand rawCommand, List<NexusCommand> commands, Map<String, NexusBlock> blocks) throws NexusFormatException {
		this.translate = new HashMap<String, String>();
		for (NexusCommand c : commands) {
			if (c.getCommandName().equalsIgnoreCase("TRANSLATE")
					&& c instanceof NexusMapOfStringsStyleCommand) {
				this.translate = ((NexusMapOfStringsStyleCommand) c).getMap();
			}
		}

		ITree tree = getDaoFactory().createTree();
		this.nodes = new HashMap<String, INode>();
		this.edges = new HashMap<Tupel<String, String>, IEdge>();
		String treeId = "";
		boolean isDefaultTree = false;

		Iterator<Token> iterator = rawCommand.getTokens().iterator();
		// optional * token
		Token token = iterator.next();
		if (token.getValue().equals("*")) {
			isDefaultTree = true;
			token = iterator.next();
		}
		// tree identifier token
		treeId = token.getValue();
		// = token
		token = iterator.next();
		if (!token.getValue().equals("=")) {
			throw new NexusFormatException("'=' sign expected at "
					+ token.getPosition());
		}
		// optionanal rooting comment token
		token = iterator.next();
		if (token.isComment()) {
			if (token.getValue().equalsIgnoreCase("[&r]")) {
				tree.setRooted(true);
			} else if (token.getValue().equalsIgnoreCase("[&u]")) {
				tree.setRooted(false);
			}
		}

		// list of tokens representing the newick representation
		List<Token> subTokens = new ArrayList<Token>();
		subTokens.add(token);

		while (iterator.hasNext()) {
			subTokens.add(iterator.next());
		}

		// compute nextUniqueId
		this.nextNodeId = 0;
		// for (String label : translate.values()){
		// if (label.matches(treeId + "(\\d+)")){
		// this.nextNodeId = Math.max(this.nextNodeId,
		// Integer.parseInt(label.substring(treeId.length())));
		// }
		// }
		// this.nextNodeId++;

		IEdge rootEdge = parseNewick(treeId, subTokens, null);
		tree.setId(treeId);
		tree.setNodes(nodes.values());
		tree.setEdges(edges.values());
		if (tree.isRooted()){
			tree.setRoot(rootEdge.getTarget());
			tree.getRoot().renumber(0);
		}
		return new NexusTreeCommand(rawCommand, treeId, tree, isDefaultTree);
	}

	private IEdge parseNewick(String treeId, List<Token> tokens, INode parent)
			throws NexusFormatException {
		INode node = getDaoFactory().createNode();
		if (node.getId().equals("")) {
			node.setId(treeId + this.nextNodeId);
		}
		this.nextNodeId++;
		int bracketCount = 0;
		Token token = null;
		List<Token> subTokens = new ArrayList<Token>();

		if (tokens.size() == 0) {
			throw new NexusFormatException(
					"can not parse empty newick string for tree " + treeId);
		}

		Iterator<Token> iterator = tokens.iterator();
		while (iterator.hasNext()) {
			token = iterator.next();
			if (bracketCount < 0) {
				throw new NexusFormatException(
						"no matching '(' before ')' at  " + token.getPosition());
			}
			subTokens.add(token);
			if (token.getValue().equals("(")) {
				if (bracketCount == 0) {
					subTokens = new ArrayList<Token>();
				}
				bracketCount++;
			} else if (token.getValue().equals(")")) {
				bracketCount--;
				if (bracketCount == 0) {
					subTokens.remove(token);
					parseNewick(treeId, subTokens, node);
					subTokens = new ArrayList<Token>();
					while (iterator.hasNext()) {
						subTokens.add(iterator.next());
					}
				}
			} else if (token.getValue().equals(",")) {
				if (bracketCount == 1) {
					subTokens.remove(token);
					parseNewick(treeId, subTokens, node);
					subTokens = new ArrayList<Token>();
				}
			}
		}
		if (bracketCount > 0) {
			throw new NexusFormatException("missing ')' at  "
					+ token.getPosition());
		}

		iterator = subTokens.iterator();
		IEdge edge = getDaoFactory().createEdge();
		edge.setSource(parent);
		edge.setTarget(node);
		while (iterator.hasNext()) {
			token = iterator.next();
			if (token.getValue().equals("(") || token.getValue().equals(")") || token.getValue().equals(",")) {
				throw new NexusFormatException(
						"node name, branch length or bootstrap value expected at "
								+ token.getPosition());
			}
			if (token.getValue().equals(":")) {
				token = iterator.next();
				try {
					edge.setLength(Double.parseDouble(token.getValue()));
				} catch (NumberFormatException e) {
					throw new NexusFormatException("branch length expected at "
							+ token.getPosition(), e);
				}
			} else if (token.isComment() && token.getValue().matches("\\d+")) {
				try {
					edge.setBootstrap(Integer.parseInt(token.getValue()));
				} catch (NumberFormatException e) {
					throw new NexusFormatException(
							"bootstap value expected at " + token.getPosition(),
							e);
				}
			} else {
				if (this.translate.containsKey(token.getValue())) {
					node.setId(this.translate.get(token.getValue()));
				} else {
					node.setId(token.getValue());
				}
			}
		}

		this.nodes.put(node.getId(), node);
		if (parent != null) {
			this.edges.put(new Tupel<String, String>(edge.getSource().getId(),
					edge.getTarget().getId()), edge);
		}
		return edge;
	}

}
